from dagster._core.storage.defs_state.base import DefsStateStorage as DefsStateStorage
from dagster._core.storage.defs_state.blob_storage_state_storage import (
    UPathDefsStateStorage as UPathDefsStateStorage,
)
